/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "terminal.pmh"

namespace fe
{

void TerminalDraw::drawPoints(const SpatialVector *vertex,
	const SpatialVector *normal,U32 vertices,
	BWORD multicolor,const Color *color)
{
	if(m_spSurfaceAccessibleI.isNull())
	{
		feLog("TerminalDraw::drawPolygons SurfaceAccessibleI is NULL\n");
		return;
	}


	//* TODO
}

void TerminalDraw::drawLines(const SpatialVector *vertex,
	const SpatialVector *normal,U32 vertices,
	StripMode strip,BWORD multicolor,
	const Color *color)
{
	drawPolygons(vertex,normal,vertices,strip,multicolor,color,TRUE);
}

void TerminalDraw::drawTriangles(const SpatialVector *vertex,
	const SpatialVector *normal,const Vector2f *texture,U32 vertices,
	StripMode strip,BWORD multicolor,const Color* color)
{
	//* TODO other strip modes

	if(strip==DrawI::e_discrete)
	{
		for(U32 m=0;m<vertices;m+=3)
		{
			if(color && multicolor)
			{
				drawPolygons(&vertex[m],&normal[m],3,DrawI::e_strip,
						multicolor,&color[m],FALSE);
			}
			else
			{
				drawPolygons(&vertex[m],&normal[m],3,DrawI::e_strip,
						multicolor,color,FALSE);
			}
		}
	}
}

void TerminalDraw::drawPolygons(const SpatialVector *vertex,
	const SpatialVector *normal,U32 vertices,
	StripMode strip,BWORD multicolor,const Color *color,BWORD openPoly)
{
	if(m_spSurfaceAccessibleI.isNull())
	{
		feLog("TerminalDraw::drawPolygons SurfaceAccessibleI is NULL\n");
		return;
	}

	//* TODO add to group
	sp<DrawMode> spDrawMode=drawMode();
	String group;
	if(spDrawMode.isValid())
	{
		group=spDrawMode->group();
	}

	sp<SurfaceAccessorI> spOutputPoint=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position);
	if(spOutputPoint.isNull())
	{
		feLog("TerminalDraw::drawPolygons points not accessible\n");
		return;
	}

	sp<SurfaceAccessorI> spOutputNormal;
	if(normal)
	{
		spOutputNormal=m_spSurfaceAccessibleI->accessor(
				SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_normal);
		if(spOutputNormal.isNull())
		{
			feLog("TerminalDraw::drawPolygons normals not accessible\n");
			return;
		}
	}

	sp<SurfaceAccessorI> spOutputVertices=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
	if(spOutputVertices.isNull())
	{
		feLog("TerminalDraw::drawPolygons vertices not accessible\n");
		return;
	}

	//* TODO e_strip (see HoudiniDraw)
	//* TODO normals, colors

	Array< Array<I32> > primVerts(1);

	Array<I32>& rVerts=primVerts[0];
	rVerts.resize(vertices);

	for(U32 index=0;index<vertices;index++)
	{
		rVerts[index]=index;
	}

	const I32 primitiveIndex=spOutputVertices->count();

	spOutputVertices->append(primVerts);
	spOutputPoint->append(vertices);

	for(U32 index=0;index<vertices;index++)
	{
		spOutputPoint->set(index,vertex[index]);
	}

	if(spOutputNormal.isValid())
	{
		for(U32 index=0;index<vertices;index++)
		{
			spOutputNormal->set(index,normal[index]);
		}
	}

	if(openPoly)
	{
		sp<SurfaceAccessorI> spOutputProperties=
			m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_properties);
		if(spOutputProperties.isValid())
		{
			spOutputProperties->set(primitiveIndex,
					SurfaceAccessibleI::e_openCurve,TRUE);
		}
	}

}

} // namespace
