/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "xgen/xgen.pmh"

#define FE_SAXG_DEBUG		TRUE

namespace fe
{

void SurfaceAccessibleXGen::reset(void)
{
	SurfaceAccessibleRecord::reset();
}

BWORD SurfaceAccessibleXGen::load(String a_filename,sp<Catalog> a_spSettings)
{
#if FE_SAXG_DEBUG
	feLog("SurfaceAccessibleXGen::load \"%s\"\n",a_filename.c_str());
#endif

	String options;
	if(a_spSettings.isValid())
	{
		options=a_spSettings->catalog<String>("options");

//		frame=a_spSettings->catalog<Real>("frame");
//		spFilter=a_spSettings->catalog< sp<Component> >("filter");
//		spContext=a_spSettings->catalogOrDefault< sp<Component> >(
//				"context",sp<Component>(NULL));
	}

	String xpdName;

	//* HACK
	xpdName="pPlane1";

	String option;
	while(!(option=options.parse()).empty())
	{
		const String property=option.parse("\"","=");
		const String value=option.parse("\"","=");

		if(property=="xpd")
		{
			xpdName=value;
		}

//		if(property=="prop")
//		{
//			variable=fe::maximum(1,value.integer());
//		}
	}

	std::vector<std::string> deltas;
	std::string response=xgapi::importPalette(a_filename.c_str(),deltas);

#if FE_SAXG_DEBUG
	feLog("SurfaceAccessibleXGen::load response \"%s\"\n",response.c_str());
#endif

	if(response.empty())
	{
		return FALSE;
	}

	const String projectPath=xgapi::getProjectPath().c_str();
	feLog("projectPath \"%s\"\n",projectPath.c_str());

	const String rootDir=xgapi::rootDir().c_str();
	feLog("rootDir \"%s\"\n",rootDir.c_str());

	String xpdPath;

	std::vector<std::string> paletteNames=xgapi::palettes();
	const I32 paletteCount=paletteNames.size();
	for(I32 paletteIndex=0;paletteIndex<paletteCount;paletteIndex++)
	{
		std::string paletteName=paletteNames[paletteIndex];
		feLog("palette %d/%d \"%s\"\n",
				paletteIndex,paletteCount,paletteName.c_str())

		XgPalette* pPalette=XgPalette::palette(paletteName);
		feLog("pPalette %p\n",pPalette);

		String dataPath;

		if(pPalette)
		{
			const XgExpression* pExpression=
					pPalette->namedAttr("xgProjectPath");
			if(pExpression)
			{
				const String project=pExpression->expression().c_str();
				feLog("project \"%s\"\n",project.c_str());
			}

			dataPath=pPalette->xgDataPathStr().c_str();
			feLog("dataPath \"%s\"\n",dataPath.c_str());
		}

		std::vector<std::string> descriptionNames=
				xgapi::descriptions(paletteName);

		const I32 descriptionCount=descriptionNames.size();
		for(I32 descriptionIndex=0;descriptionIndex<descriptionCount;
				descriptionIndex++)
		{
			std::string descriptionName=descriptionNames[descriptionIndex];
			feLog("  description %d/%d \"%s\"\n",
					descriptionIndex,descriptionCount,descriptionName.c_str())

			XgDescription* pDescription=
					pPalette? pPalette->description(descriptionName): NULL;

			dataPath=xgapi::expandFilepath(
					dataPath.c_str(),descriptionName).c_str();
			feLog("    expanded dataPath \"%s\"\n",dataPath.c_str());

			const String bakePath=dataPath+descriptionName.c_str()+"/Bake";
			xpdPath=bakePath+"/"+xpdName+".xpd";

			XgPrimitive* pPrimitive=
					pDescription? pDescription->activePrimitive(): NULL;

			feLog("    primitive %p\n",pPrimitive);

			std::vector<std::string> objectNames=
					xgapi::objects(paletteName,descriptionName);

			const I32 objectCount=objectNames.size();
			for(I32 objectIndex=0;objectIndex<objectCount;
					objectIndex++)
			{
				std::string objectName=objectNames[objectIndex];
				feLog("    object %d/%d \"%s\"\n",
						objectIndex,objectCount,objectName.c_str())

				std::vector<std::string> attrNames=
						xgapi::attrs(paletteName,descriptionName,objectName);

				const I32 attrCount=attrNames.size();
#if FALSE
				for(I32 attrIndex=0;attrIndex<attrCount;
						attrIndex++)
				{
					std::string attrName=attrNames[attrIndex];
					feLog("      attr %d/%d \"%s\"\n",
							attrIndex,attrCount,attrName.c_str())
				}
#else
				feLog("      %d attributes\n",attrCount);
#endif
			}
		}
	}

	//* using Record writing code from SurfaceAccessibleLua

	sp<Scope> spScope=registry()->create("Scope");
	bind(spScope);

	sp<Layout> spLayoutPoint=spScope->declare("pt");
	spLayoutPoint->populate("spc:at","vector3");

	sp<Layout> spLayoutPrimitive=spScope->declare("prim");
	spLayoutPrimitive->populate("surf:verts","intarray");
	spLayoutPrimitive->populate("properties","int");

	Accessor<SpatialVector> aPosition;
	aPosition.setup(spScope,"spc:at");

	Accessor< Array<I32> > aVerts;
	aVerts.setup(spScope,"surf:verts");

	Accessor<I32> aProperties;
	aProperties.setup(spScope,"properties");

	SurfaceModel surfaceModelRV;
	surfaceModelRV.bind(record());

	sp<RecordGroup> spPointRG=surfaceModelRV.surfacePointRG();
	sp<RecordGroup> spPrimRG=surfaceModelRV.surfacePrimitiveRG();

	feLog("xpdPath \"%s\"\n",xpdPath.c_str());

	XpdFile xpd;
	if(!xpd.openRead(xpdPath.c_str()))
	{
		feLog("xpd open failed\n");
		return FALSE;
	}

	const I32 numCVs=xpd.numCVs();
	feLog("numCVs %d\n",numCVs);

	I32 pointIndex=0;
	I32 faceIndex=0;
	while (xpd.nextFace())
	{
		int blockIndex = xpd.blockIndex("BakedGroom");
		if(!xpd.findBlock(blockIndex))
		{
			feLog("xpd has no BakedGroom\n");
			return FALSE;
		}

		I32 subPrimIndex=0;

		safevector<float> data;
		while(xpd.readPrim(data))
		{
			Record primitiveRecord=spScope->createRecord(spLayoutPrimitive);
			spPrimRG->add(primitiveRecord);

			Array<I32>& rVertArray=aVerts(primitiveRecord);

			aProperties(primitiveRecord)=TRUE;

			feLog("size %d\n",data.size());

			const I32 subCount=numCVs;
			for(I32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const float* float3=
						static_cast<float *>((void*)(&data[3+3*subIndex]));
				const SpatialVector point(float3[0],float3[1],float3[2]);
				feLog("face %d prim %d cv %d point %s\n",
						faceIndex,subPrimIndex,subIndex,c_print(point));

				Record pointRecord=spScope->createRecord(spLayoutPoint);
				spPointRG->add(pointRecord);

				aPosition(pointRecord)=point;

				rVertArray.push_back(pointIndex++);
			}

			subPrimIndex++;
		}
		faceIndex++;
	}

	return TRUE;
}

} // namespace
