/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "data/data.pmh"

namespace fe
{

FE_DL_EXPORT Library* CreateDataLibrary(sp<Master> spMaster)
{
	if(!spMaster->catalog()->cataloged("path:media"))
	{
		fe::String mediaPath=fe::String("${FE_MEDIA}").substituteEnv();
		if(mediaPath.empty())
		{
			fe::String rootPath=fe::String("${FE_ROOT}").substituteEnv();

			rootPath="";

			if(rootPath.empty())
			{
				rootPath=".";

				const String loadPath=System::getLoadPath(
						(void*)CreateDataLibrary,TRUE);

				if(!loadPath.empty())
				{
					rootPath=loadPath;
				}
			}

			mediaPath=rootPath+"/media";
		}
		spMaster->catalog()->catalog<fe::String>("path:media")=
				mediaPath.substituteEnv();

		const String verbose=System::getVerbose();
		if(verbose=="all")
		{
			feLog("setting 'path:media' to \"%s\"\n",mediaPath.c_str());
		}
	}

	fe::Library *pLibrary = new fe::Library();
	pLibrary->setName("default_data");
	pLibrary->add<fe::SegmentStore>("StoreI.SegmentStore.fe");
	pLibrary->add<fe::Scope>("Scope.Scope.fe");
	pLibrary->add<fe::data::AsciiFileStream>("FileStreamI.Ascii.fe");
	pLibrary->add<fe::data::BinaryFileStream>("FileStreamI.Binary.fe");

	pLibrary->add<fe::AsSystem>("PopulateI.AsSystem.fe");
	pLibrary->add<fe::AsSequenceSignal>("PopulateI.AsSequenceSignal.fe");
	pLibrary->addSingleton<fe::RecordCookbook>("Component.RecordCookbook.fe");
	return pLibrary;

}

}

