/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "plugin.pmh"

namespace fe
{

Catalog::Catalog(void):
	m_highest(0)
{
}

Catalog::~Catalog(void)
{
	catalogClear();
}

sp<TypeMaster> Catalog::typeMaster(void) const
{
	return registry()->master()->typeMaster();
}

void Catalog::catalogDump(void)
{
	feLog("Catalog:\n");
	for(std::map< String, InstanceMap >::iterator it1=m_instanceMapMap.begin();
			it1!=m_instanceMapMap.end();it1++)
	{
		for(std::map< String, Instance >::iterator it2=it1->second.begin();
				it2!=it1->second.end();it2++)
		{
			feLog(" \"%s\" \"%s\" 0x%x \"%s\"\n", it1->first.c_str(),
					it2->first.c_str(), it2->second.data(),
					c_print(it2->second));
		}
	}
	feLog("\n");
}

void Catalog::catalogKeys(Array<String> &a_keys)
{
	for(std::map< I32, String >::iterator it = m_orderMap.begin();
			it != m_orderMap.end(); it++)
	{
		a_keys.push_back(it->second);
	}
}

void Catalog::catalogProperties(const String a_name,
	Array<String> &a_properties)
{
	std::map<String, InstanceMap>::iterator it1 = m_instanceMapMap.find(a_name);
	if(it1 == m_instanceMapMap.end())
	{
		return;
	}

	for(std::map< String, Instance >::iterator it2=it1->second.begin();
			it2!=it1->second.end();it2++)
	{
		a_properties.push_back(it2->first);
	}
}

//* TODO also build a valid m_orderMap
void Catalog::catalogOverlay(sp<Catalog> a_spOtherCatalog,
	Array<String>* a_pTypeList)
{
	for(std::map< String, InstanceMap >::iterator it=
			a_spOtherCatalog->m_instanceMapMap.begin();
			it != a_spOtherCatalog->m_instanceMapMap.end(); it++)
	{
		const String key=it->first;

		for(std::map< String, Instance >::iterator it2=it->second.begin();
				it2!=it->second.end();it2++)
		{
			const String property=it2->first;
			const Instance instance = it2->second;

			if(a_pTypeList)
			{
				const String typeName=instance.type()->typeinfo().ref().name();

				const U32 nameCount=a_pTypeList->size();
				U32 nameIndex=0;
				for(;nameIndex<nameCount;nameIndex++)
				{
//					feLog("type \"%s\" vs \"%s\"\n",typeName.c_str(),
//							(*a_pTypeList)[nameIndex].c_str());

					if(typeName==(*a_pTypeList)[nameIndex])
					{
						break;
					}
				}
				if(nameIndex==nameCount)
				{
					continue;
				}
			}

			//* make an independent instance
			catalogInstance(key,property).create(instance.type());
			catalogInstance(key,property).assign(instance);
		}
	}
}

bool Catalog::catalogLookup(const String a_name, const String a_property,
	Instance &a_instance)
{
	std::map<String, InstanceMap>::iterator it1 = m_instanceMapMap.find(a_name);
	if(it1 == m_instanceMapMap.end())
	{
		return false;
	}
	std::map<String, Instance>::iterator it2 =
			m_instanceMapMap[a_name].find(a_property);
	if(it2 == m_instanceMapMap[a_name].end())
	{
		return false;
	}
	a_instance = it2->second;
	return true;
}

} /* namespace */
